# coding=utf-8
import pytest

from insights.combiners.httpd_conf import HttpdConfTree, HttpdConfSclHttpd24Tree, HttpdConfSclJbcsHttpd24Tree
from insights.parsers import ParseException, SkipException, httpd_conf
from insights.tests import context_wrap

HTTPD_CONF_MIXED_NAME = '''
H2Push on
'''

HTTPD_CONF_MIXED = '''
JustFotTest_NoSec "/var/www/cgi"
# prefork MPM
<IfModule prefork.c>
ServerLimit      256
ThreadsPerChild  16
JustForTest      "AB"
MaxClients       256
</IfMoDuLe>

IncludeOptional conf.d/*.conf
'''.strip()

HTTPD_CONF_CONTINUATION = r'''
JustFotTest_NoSec "/var/www/cgi"
CustomLog logs/ssl_request_log \
"%t %h %{SSL_PROTOCOL}x %{SSL_CIPHER}x \"%r\" %b"
# prefork MPM
<IfModule prefork.c>
ServerLimit      256
ThreadsPerChild  16
JustForTest      "AB"
MaxClients       256
</IfModule>

IncludeOptional conf.d/*.conf
'''.strip()

HTTPD_CONF_1 = '''
JustFotTest_NoSec "/var/www/cgi"
# prefork MPM
<IfModule prefork.c>
ServerLimit      256
ThreadsPerChild  16
JustForTest      "AB"
MaxClients       256
</IfModule>

IncludeOptional conf.d/*.conf
'''.strip()

HTTPD_CONF_2 = '''
JustForTest_NoSec "/var/www/cgi"
# prefork MPM
<IfModule prefork.c>
ServerLimit      1024
JustForTest      "ABC"
MaxClients       1024
</IfModule>
'''.strip()

HTTPD_CONF_3 = '''
# prefork MPM
<IfModule prefork.c>
ServerLimit      256
MaxClients       512
</IfModule>
'''.strip()

HTTPD_CONF_SHADOWTEST_1 = '''
Foo 1A
Foo 1B
Foo 1C
<IfModule prefork.c>
Foo 1xA
Foo 1xB
Foo 1xC
Bar 1A
Bar 1B
Bar 1C
</IfModule>

IncludeOptional conf.d/*.conf
'''.strip()

HTTPD_CONF_SHADOWTEST_2 = '''
Foo 2A
Foo 2B
Foo 2C
<IfModule ASDF.prefork.c.ASDF>
Foo 2xA
Foo 2xB
Foo 2xC
Bar 2A
Bar 2B
Bar 2C
</IfModule>
'''.strip()

HTTPD_CONF_SHADOWTEST_3 = '''
Foo 3A
Foo 3B
Foo 3C
<IfModule prefork.c>
Foo 3xA
Foo 3xB
Foo 3xC
Bar 3A
Bar 3B
Bar 3C
</IfModule>
'''.strip()

HTTPD_CONF_MAIN_1 = '''
ServerRoot "/etc/httpd"
Listen 80

# Load config files in the "/etc/httpd/conf.d" directory, if any.
IncludeOptional conf.d/*.conf
SSLProtocol -ALL +TLSv1.2  # SSLv3
'''.strip()

HTTPD_CONF_MAIN_2 = '''
# Load config files in the "/etc/httpd/conf.d" directory, if any.
IncludeOptional conf.d/*.conf

ServerRoot "/etc/httpd"
Listen 80
'''.strip()

HTTPD_CONF_MAIN_3 = '''
ServerRoot "/etc/httpd"

# Load config files in the "/etc/httpd/conf.d" directory, if any.
IncludeOptional conf.d/*.conf
IncludeOptional conf.d/*/*.conf

Listen 80
'''.strip()

HTTPD_CONF_MAIN_4 = '''
IncludeOptional conf.d/*.conf
IncludeOptional conf.modules.d/*.conf
Listen 80
'''.strip()

HTTPD_CONF_FILE_1 = '''
ServerRoot "/home/skontar/httpd"
Listen 8080
'''.strip()

HTTPD_CONF_FILE_2 = '''
ServerRoot "/home/skontar/www"
'''.strip()

HTTPD_CONF_FILE_3 = '''
LoadModule access_compat_module modules/mod_access_compat.so
LoadModule actions_module modules/mod_actions.so
LoadModule alias_module modules/mod_alias.so
LoadModule mpm_prefork_module modules/mod_mpm_prefork.so
'''.strip()

HTTPD_CONF_MORE = '''
UserDir disable
UserDir enable bob
'''.strip()

HTTPD_EMPTY_LAST = '''
#
# Directives controlling the display of server-generated directory listings.
#
# Required modules: mod_authz_core, mod_authz_host,
#                   mod_autoindex, mod_alias
#
# To see the listing of a directory, the Options directive for the
# directory must include "Indexes", and the directory must not contain
# a file matching those listed in the DirectoryIndex directive.
#

#
# IndexOptions: Controls the appearance of server-generated directory
# listings.
#
IndexOptions FancyIndexing HTMLTable VersionSort

# We include the /icons/ alias for FancyIndexed directory listings.  If
# you do not use FancyIndexing, you may comment this out.
#
Alias /icons/ "/usr/share/httpd/icons/"

<Directory "/usr/share/httpd/icons"> 
    Options Indexes MultiViews FollowSymlinks
    AllowOverride None
    Require all granted
</Directory>

#
# AddIcon* directives tell the server which icon to show for different
# files or filename extensions.  These are only displayed for
# FancyIndexed directories.
#
AddIconByEncoding (CMP,/icons/compressed.gif) x-compress x-gzip

AddIconByType (TXT,/icons/text.gif) text/*
AddIconByType (IMG,/icons/image2.gif) image/*
AddIconByType (SND,/icons/sound2.gif) audio/*
AddIconByType (VID,/icons/movie.gif) video/*

AddIcon /icons/binary.gif .bin .exe
AddIcon /icons/binhex.gif .hqx
AddIcon /icons/tar.gif .tar
AddIcon /icons/world2.gif .wrl .wrl.gz .vrml .vrm .iv
AddIcon /icons/compressed.gif .Z .z .tgz .gz .zip
AddIcon /icons/a.gif .ps .ai .eps
AddIcon /icons/layout.gif .html .shtml .htm .pdf
AddIcon /icons/text.gif .txt
AddIcon /icons/c.gif .c
AddIcon /icons/p.gif .pl .py
AddIcon /icons/f.gif .for
AddIcon /icons/dvi.gif .dvi
AddIcon /icons/uuencoded.gif .uu
AddIcon /icons/script.gif .conf .sh .shar .csh .ksh .tcl
AddIcon /icons/tex.gif .tex
AddIcon /icons/bomb.gif core.

AddIcon /icons/back.gif ..
AddIcon /icons/hand.right.gif README
AddIcon /icons/folder.gif ^^DIRECTORY^^
AddIcon /icons/blank.gif ^^BLANKICON^^

#
# DefaultIcon is which icon to show for files which do not have an icon
# explicitly set.
#
DefaultIcon /icons/unknown.gif

#
# AddDescription allows you to place a short description after a file in
# server-generated indexes.  These are only displayed for FancyIndexed
# directories.
# Format: AddDescription "description" filename
#
#AddDescription "GZIP compressed document" .gz
#AddDescription "tar archive" .tar
#AddDescription "GZIP compressed tar archive" .tgz

#
# ReadmeName is the name of the README file the server will look for by
# default, and append to directory listings.
#
# HeaderName is the name of a file which should be prepended to
# directory indexes.
ReadmeName README.html
HeaderName HEADER.html

#
# IndexIgnore is a set of filenames which directory indexing should ignore
# and not include in the listing.  Shell-style wildcarding is permitted.
#
IndexIgnore .??* *~ *# HEADER* README* RCS CVS *,v *,t

'''.lstrip()  # noqa W291


HTTPD_CONF_NEST_1 = """
<VirtualHost 128.39.140.28>
    <Directory /var/www/example>
        Options FollowSymLinks
        AllowOverride None
    </Directory>
    <IfModule mod_php4.c>
        php_admin_flag safe_mode Off
        php_admin_value register_globals    0
    </IfModule>
    DirectoryIndex index.php
    <IfModule mod_rewrite.c>
        RewriteEngine On
        RewriteRule .* /index.php
    </IfModule>
    <IfModule mod_rewrite.c>
        RewriteEngine Off
    </IfModule>
    <IfModule !php5_module>
        <IfModule !php4_module>
            <FilesMatch ".php[45]?$">
                Order allow,deny
                Deny from all
            </FilesMatch>
            <FilesMatch ".php[45]?$">
                Order deny,allow
            </FilesMatch>
        </IfModule>
    </IfModule>
    DocumentRoot /var/www/example
    ServerName www.example.com
    ServerAlias admin.example.com
</VirtualHost>
<IfModule !php5_module>
  <IfModule !php4_module>
    <Location />
        <FilesMatch ".php[45]">
            Order allow,deny
            Deny from all
        </FilesMatch>
    </Location>
  </IfModule>
</IfModule>
<IfModule mod_rewrite.c>
    RewriteEngine Off
</IfModule>
LogLevel warn
DocumentRoot "/var/www/html_cgi"
IncludeOptional conf.d/*.conf
EnableSendfile on
""".strip()

HTTPD_CONF_NEST_2 = """
DocumentRoot "/var/www/html"
<VirtualHost 128.39.140.30>
    <IfModule !php5_module>
        <IfModule !php4_module>
            <FilesMatch ".php[45]?$">
                Order allow,deny
                Deny from all
            </FilesMatch>
            <FilesMatch ".php[45]?$">
                Order deny,allow
            </FilesMatch>
        </IfModule>
    </IfModule>
    DocumentRoot /var/www/example1
    ServerName www.example1.com
    ServerAlias admin.example1.com
</VirtualHost>
<IfModule !php5_module>
  <IfModule !php4_module>
    <Location />
        <FilesMatch test>
            Order deny,allow
            Allow from all
        </FilesMatch>
        <FilesMatch ".php[45]">
            Order deny,allow
        </FilesMatch>
    </Location>
  </IfModule>
</IfModule>
<IfModule mod_rewrite.c>
    RewriteEngine On
</IfModule>
EnableSendfile off
""".strip()

HTTPD_CONF_NEST_3 = """
<VirtualHost 128.39.140.28>
    <IfModule !php5_module>
        Testphp php5_v3_1
        <IfModule !php4_module>
            Testphp php4_v3_1
        </IfModule>
        Testphp php5_v3_2
    </IfModule>
</VirtualHost>
<IfModule !php5_module>
  Testphp php5_3_a
  <IfModule !php4_module>
    Testphp php4_3_a
  </IfModule>
</IfModule>
""".strip()

HTTPD_CONF_NEST_4 = """
<VirtualHost 128.39.140.30>
    <IfModule !php5_module>
        Testphp php5_v4_1
        <IfModule !php4_module>
            Testphp php4_v4_1
        </IfModule>
        Testphp php5_v4_2
    </IfModule>
</VirtualHost>
<IfModule !php5_module>
  Testphp php5_4_b
  <IfModule !php4_module>
    Testphp php4_4_b
  </IfModule>
</IfModule>
""".strip()


HTTPD_REGEX_AND_OP_ATTRS = r"""
RewriteCond %{HTTP:Accept-Encoding} \b(x-)?gzip\b
RedirectMatch ^\/?pulp_puppet\/forge\/[^\/]+\/[^\/]+\/(?!api\/v1\/releases\.json)(.*)$ /$1
RewriteCond %1%2 (^|&|;)([^(&|;)].*|$)
RewriteCond %{HTTP:Accept-Encoding} \b(x-)?gzip\b
<IfVersion < 2.4>
    Allow from all
  </IfVersion>
""".strip()


HTTPD_EMBEDDED_QUOTES = r"""
# DirectoryIndex: sets the file that Apache will serve if a directory
# is requested.
#
<IfModule dir_module>
    DirectoryIndex index.html
</IfModule>

#
# The following lines prevent .htaccess and .htpasswd files from being
# viewed by Web clients.
#
<Files ".ht*">
    Require all denied
</Files>

#
# ErrorLog: The location of the error log file.
# If you do not specify an ErrorLog directive within a <VirtualHost>
# container, error messages relating to that virtual host will be
# logged here.  If you *do* define an error logfile for a <VirtualHost>
# container, that host's errors will be logged there and not here.
#
ErrorLog "logs/error_log"

  RequestHeader   whatever # last value is taken
  
  
  RequestHeader   unset   Proxy

#
# LogLevel: Control the number of messages logged to the error_log.
# Possible values include: debug, info, notice, warn, error, crit,
# alert, emerg.
#
LogLevel warn

<IfModule log_config_module>
    #
    # The following directives define some format nicknames for use with
    # a CustomLog directive (see below).
    #
    LogFormat "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\"" combined
    LogFormat "%h %l %u %t \"%r\" %>s %b" common

    <IfModule logio_module>
      # You need to enable mod_logio.c to use %I and %O
      LogFormat "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\" %I %O" combinedio
    </IfModule>

    #
    # The location and format of the access logfile (Common Logfile Format).
    # logged therein and *not* in this file.
    #
    #CustomLog "logs/access_log" common

    #
    # If you prefer a logfile with access, agent, and referer information
    # (Combined Logfile Format) you can use the following directive.
    #
    CustomLog "logs/access_log" combined
</IfModule>


DNSSDEnable on
#DNSSDAutoRegisterVHosts on
#DNSSDAutoRegisterUserDir on
""".strip()  # noqa W293


UNICODE_COMMENTS = """
#Alterações realizadas por issue no Insights
DNSSDEnable on
#DNSSDAutoRegisterVHosts on
#DNSSDAutoRegisterUserDir on
"""


MULTIPLE_INCLUDES = """
<IfVersion < 2.4>
  Include /etc/httpd/conf.d/05-foreman.d/*.conf
</IfVersion>
<IfVersion >= 2.4>
  IncludeOptional /etc/httpd/conf.d/05-foreman.d/*.conf
</IfVersion>
"""

HTTPD_EMPTY_ATTR = """
<VirtualHost *:80>
    RequestHeader set X_FORWARDED_PROTO "http"
    RequestHeader set SSL_CLIENT_S_DN ""
    RequestHeader set SSL_CLIENT_CERT ""
    RequestHeader set SSL_CLIENT_VERIFY ""
</VirtualHost>
""".strip()

HTTPD_MULTILINE_QUOTE = r"""
#LoadModule ldap_module modules/mod_ldap.so
#LoadModule auth_ldap_module modules/mod_auth_ldap.so

LDAPVerifyServerCert Off
#LogLevel debug

Listen 81
<VirtualHost *:81>
    SetEnv GIT_PROJECT_ROOT /var/www/git
    SetEnv GIT_HTTP_EXPORT_ALL
    DocumentRoot /var/www
#    ScriptAlias /git/ /usr/libexec/git-core/git-http-backend/
    ProxyPass "http://backend.example.com/a-long-path-to-demonstrate" \
        connectiontimeout=1 retry=0
    ScriptAliasMatch \
        "(?x)^/git/(.*/(HEAD | \
        info/refs | \
        objects/(info/[^/]+ | \
        [0-9a-f]{2}/[0-9a-f]{38} | \
        pack/pack-[0-9a-f]{40}\.(pack|idx)) | \
        git-(upload|receive)-pack))$" \
        "/usr/libexec/git-core/git-http-backend/$1"

</VirtualHost>
"""

INVALID_CONF = """
<<<>>>><>KDLKJLDSF><SDNF<KNSD><FN
""".strip()


def test_mixed_case_tags():
    httpd = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_MIXED, path='/etc/httpd/conf/httpd.conf'))
    assert httpd.find("ServerLimit").value == 256


def test_line_continuation():
    httpd = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_CONTINUATION, path='/etc/httpd/conf/httpd.conf'))
    val = httpd.find("CustomLog")[0].attrs
    assert val == [r'logs/ssl_request_log', r'%t %h %{SSL_PROTOCOL}x %{SSL_CIPHER}x "%r" %b'], val


def test_nopath():
    # no path
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_2))
    try:
        HttpdConfTree([httpd2])
        # assert result['IfModule', 'prefork.c']['ServerLimit'][-1].value == 1024
        exception_happened = False
    except:
        exception_happened = True
    assert exception_happened

    # no httpd.conf
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_2, path='/etc/httpd/conf.d/00-z.conf'))
    try:
        HttpdConfTree([httpd2])
        # assert result['IfModule', 'prefork.c']['ServerLimit'][-1].value == 1024
        exception_happened = False
    except:
        exception_happened = True
    assert exception_happened

    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_2, path='/etc/httpd/conf/httpd.conf'))
    result = HttpdConfTree([httpd2])
    assert result['IfModule', 'prefork.c']['ServerLimit'][-1].value == 1024

    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_2, path='/laaalalalala/blablabla/httpd.conf'))
    result = HttpdConfTree([httpd2])
    assert result['IfModule', 'prefork.c']['ServerLimit'][-1].value == 1024

    # no include in httpd.conf
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_2, path='/etc/httpd/conf/httpd.conf'))
    httpd3 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_3, path='/etc/httpd/conf.d/z-z.conf'))
    result = HttpdConfTree([httpd2, httpd3])
    assert result['IfModule', 'prefork.c']['ServerLimit'][-1].value == 1024

    # no include in httpd.conf
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_2, path='/etc/httpd/conf/httpd.conf'))
    httpd3 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_3, path='/etc/httpd/conf.d/aaa.conf'))
    result = HttpdConfTree([httpd3, httpd2])
    assert len(result['IfModule', 'prefork.c']['ServerLimit']) == 1
    assert result['IfModule', 'prefork.c']['ServerLimit'][0].value == 1024
    assert result['IfModule', 'prefork.c']['ServerLimit'][-1].value == 1024

    # with an include
    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_1, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_2, path='/etc/httpd/conf.d/00-z.conf'))
    result = HttpdConfTree([httpd1, httpd2])
    assert len(result['IfModule', 'prefork.c']['ServerLimit']) == 2
    assert result['IfModule', 'prefork.c']['ServerLimit'][0].value == 256
    assert result['IfModule', 'prefork.c']['ServerLimit'][-1].value == 1024

    # colliding filenames
    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_1, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_2, path='/etc/httpd/conf.d/00-z.conf'))
    httpd3 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_3, path='/etc/httpd/conf.d/00-z.conf'))
    result = HttpdConfTree([httpd1, httpd2, httpd3])
    assert len(result['IfModule', 'prefork.c']['ServerLimit']) == 3
    assert result['IfModule', 'prefork.c']['ServerLimit'][0].value == 256  # httpd1
    assert result['IfModule', 'prefork.c']['ServerLimit'][1].value == 1024     # httpd2
    assert result['IfModule', 'prefork.c']['ServerLimit'][-1].value == 256   # httpd3
    assert len(result['IfModule', 'prefork.c']['MaxClients']) == 3
    assert result['IfModule', 'prefork.c']['MaxClients'][0].value == 256  # httpd1
    assert result['IfModule', 'prefork.c']['MaxClients'][1].value == 1024     # httpd2
    assert result['IfModule', 'prefork.c']['MaxClients'][-1].value == 512   # httpd3

    # testing other ways to access the same indices
    assert result['IfModule', 'prefork.c']['MaxClients'][0].value == 256   # httpd1
    assert result['IfModule', 'prefork.c']['MaxClients'][2].value == 512   # httpd3
    assert result['IfModule', 'prefork.c']['MaxClients'][-1].value == 512  # httpd3


def test_active_httpd():
    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_1, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_2, path='/etc/httpd/conf.d/00-z.conf'))
    httpd3 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_3, path='/etc/httpd/conf.d/z-z.conf'))
    result = HttpdConfTree([httpd1, httpd2, httpd3])

    assert result['IfModule', 'prefork.c']['MaxClients'][-1].value == 512
    assert result['IfModule', 'prefork.c']['MaxClients'][-1].file_path == '/etc/httpd/conf.d/z-z.conf'
    assert result['IfModule', 'prefork.c']['ThreadsPerChild'][-1].value == 16
    assert result['IfModule', 'prefork.c']['ServerLimit'][-1].value == 256
    assert result['IfModule', 'prefork.c']['JustForTest'][-1].file_name == '00-z.conf'
    assert result['JustForTest_NoSec'][0].line == 'JustForTest_NoSec "/var/www/cgi"'


def test_shadowing():
    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_SHADOWTEST_1, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_SHADOWTEST_2, path='/etc/httpd/conf.d/00-z.conf'))
    httpd3 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_SHADOWTEST_3, path='/etc/httpd/conf.d/z-z.conf'))
    result = HttpdConfTree([httpd1, httpd2, httpd3])

    assert len(result["Foo"]) == 9
    assert len(result["IfModule"]) == 3
    assert len(result["IfModule"]["Foo"]) == 9
    assert len(result["IfModule"]["Bar"]) == 9
    assert len(result["IfModule", "prefork.c"]) == 2
    assert len(result["IfModule", "prefork.c"]["Foo"]) == 6
    assert len(result["IfModule", "prefork.c"]["Bar"]) == 6
    assert len(result["IfModule", "prefork.c"][0]["Foo"]) == 3
    assert len(result["IfModule", "prefork.c"][1]["Foo"]) == 3
    assert len(result["IfModule", "prefork.c"][0]["Bar"]) == 3
    assert len(result["IfModule", "prefork.c"][1]["Bar"]) == 3


def test_splits():
    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_MAIN_1, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_FILE_1, path='/etc/httpd/conf.d/00-a.conf'))
    httpd3 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_FILE_2, path='/etc/httpd/conf.d/01-b.conf'))
    result = HttpdConfTree([httpd1, httpd2, httpd3])

    server_root = result['ServerRoot'][-1]
    assert server_root.value == '/home/skontar/www'
    assert server_root.line == 'ServerRoot "/home/skontar/www"'
    assert server_root.file_name == '01-b.conf'
    assert server_root.file_path == '/etc/httpd/conf.d/01-b.conf'

    listen = result["Listen"][-1]
    assert listen.value == 8080
    assert listen.line == 'Listen 8080'
    assert listen.file_name == '00-a.conf'
    assert listen.file_path == '/etc/httpd/conf.d/00-a.conf'

    ssl_proto = result['SSLProtocol'][-1]
    assert ssl_proto.attrs == ["-ALL", "+TLSv1.2", "#", "SSLv3"]

    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_MAIN_2, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_FILE_1, path='/etc/httpd/conf.d/00-a.conf'))
    httpd3 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_FILE_2, path='/etc/httpd/conf.d/01-b.conf'))
    result = HttpdConfTree([httpd1, httpd2, httpd3])

    server_root = result['ServerRoot'][-1]
    assert server_root.value == '/etc/httpd'
    assert server_root.line == 'ServerRoot "/etc/httpd"'
    assert server_root.file_name == 'httpd.conf'
    assert server_root.file_path == '/etc/httpd/conf/httpd.conf'

    listen = result["Listen"][-1]
    assert listen.value == 80
    assert listen.line == 'Listen 80'
    assert listen.file_name == 'httpd.conf'
    assert listen.file_path == '/etc/httpd/conf/httpd.conf'

    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_MAIN_3, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_FILE_1, path='/etc/httpd/conf.d/00-a.conf'))
    httpd3 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_FILE_2, path='/etc/httpd/conf.d/01-b.conf'))
    result = HttpdConfTree([httpd1, httpd2, httpd3])

    server_root = result['ServerRoot'][-1]
    assert server_root.value == '/home/skontar/www'
    assert server_root.line == 'ServerRoot "/home/skontar/www"'
    assert server_root.file_name == '01-b.conf'
    assert server_root.file_path == '/etc/httpd/conf.d/01-b.conf'
    assert listen.value == 80
    assert listen.line == 'Listen 80'
    assert listen.file_name == 'httpd.conf'
    assert listen.file_path == '/etc/httpd/conf/httpd.conf'


def test_httpd_one_file_overwrites():
    httpd = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_MORE, path='/etc/httpd/conf/httpd.conf'))
    result = HttpdConfTree([httpd])

    active_setting = result['UserDir'][-1]
    assert active_setting.value == 'enable bob'
    assert active_setting.file_path == '/etc/httpd/conf/httpd.conf'
    assert active_setting.file_name == 'httpd.conf'
    assert active_setting.line == 'UserDir enable bob', active_setting.line

    setting_list = result['UserDir']
    assert len(setting_list) == 2
    assert setting_list[0].value == 'disable'
    assert setting_list[0].line == 'UserDir disable'
    assert setting_list[0].file_path == '/etc/httpd/conf/httpd.conf'
    assert setting_list[0].file_name == 'httpd.conf'
    assert setting_list[0].section is None
    assert setting_list[1].value == 'enable bob'
    assert setting_list[1].line == 'UserDir enable bob'
    assert setting_list[1].file_path == '/etc/httpd/conf/httpd.conf'
    assert setting_list[1].file_name == 'httpd.conf'
    assert setting_list[1].section_name is None


def test_httpd_conf_empty():
    with pytest.raises(SkipException):
        assert httpd_conf.HttpdConf(context_wrap('', path='/etc/httpd/httpd.conf')) is None


def test_httpd_conf_tree_with_load_modules():
    httpd1 = httpd_conf.HttpdConfSclHttpd24(context_wrap(HTTPD_CONF_MAIN_4, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConfSclHttpd24(context_wrap(HTTPD_CONF_MORE, path='/etc/httpd/conf.d/01-b.conf'))
    httpd3 = httpd_conf.HttpdConfSclHttpd24(context_wrap(HTTPD_CONF_FILE_3, path='/etc/httpd/conf.modules.d/02-c.conf'))
    result = HttpdConfTree([httpd1, httpd2, httpd3])
    userdirs = result['UserDir']
    assert len(userdirs) == 2
    assert userdirs[-1].value == 'enable bob'
    load_module_list = result['LoadModule']
    assert len(load_module_list) == 4
    assert result['LoadModule'][3].value == 'mpm_prefork_module modules/mod_mpm_prefork.so'
    assert result['LoadModule'][3].file_path == '/etc/httpd/conf.modules.d/02-c.conf'


def test_httpd_conf_scl_httpd24_tree():
    httpd1 = httpd_conf.HttpdConfSclHttpd24(
        context_wrap(HTTPD_CONF_MAIN_4, path='/opt/rh/httpd24/root/etc/httpd/conf/httpd.conf')
    )
    httpd2 = httpd_conf.HttpdConfSclHttpd24(
        context_wrap(HTTPD_CONF_MORE, path='/opt/rh/httpd24/root/etc/httpd/conf.d/01-b.conf')
    )
    httpd3 = httpd_conf.HttpdConfSclHttpd24(
        context_wrap(HTTPD_CONF_FILE_3, path='/opt/rh/httpd24/root/etc/httpd/conf.modules.d/02-c.conf')
    )
    result = HttpdConfSclHttpd24Tree([httpd1, httpd2, httpd3])
    userdirs = result['UserDir']
    assert len(userdirs) == 2
    assert userdirs[-1].value == 'enable bob'
    load_module_list = result['LoadModule']
    assert len(load_module_list) == 4
    assert result['LoadModule'][3].value == 'mpm_prefork_module modules/mod_mpm_prefork.so'
    assert result['LoadModule'][3].file_path == '/opt/rh/httpd24/root/etc/httpd/conf.modules.d/02-c.conf'


def test_httpd_conf_jbcs_httpd24_tree():
    httpd1 = httpd_conf.HttpdConfSclJbcsHttpd24(
        context_wrap(HTTPD_CONF_MAIN_4, path='/opt/rh/jbcs-httpd24/root/etc/httpd/conf/httpd.conf')
    )
    httpd2 = httpd_conf.HttpdConfSclJbcsHttpd24(
        context_wrap(HTTPD_CONF_MORE, path='/opt/rh/jbcs-httpd24/root/etc/httpd/conf.d/01-b.conf')
    )
    httpd3 = httpd_conf.HttpdConfSclJbcsHttpd24(
        context_wrap(HTTPD_CONF_FILE_3, path='/opt/rh/jbcs-httpd24/root/etc/httpd/conf.modules.d/02-c.conf')
    )
    result = HttpdConfSclJbcsHttpd24Tree([httpd1, httpd2, httpd3])
    userdirs = result['UserDir']
    assert len(userdirs) == 2
    assert userdirs[-1].value == 'enable bob'
    load_module_list = result['LoadModule']
    assert len(load_module_list) == 4
    assert result['LoadModule'][3].value == 'mpm_prefork_module modules/mod_mpm_prefork.so'
    assert result['LoadModule'][3].file_path == '/opt/rh/jbcs-httpd24/root/etc/httpd/conf.modules.d/02-c.conf'


def test_httpd_nested_conf_file():
    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_MAIN_3, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_FILE_1, path='/etc/httpd/conf.d/00-a.conf'))
    httpd3 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_FILE_2, path='/etc/httpd/conf.d/d1/hello.conf'))
    result = HttpdConfTree([httpd1, httpd2, httpd3])

    server_root = result['ServerRoot'][-1]
    assert server_root.value == '/home/skontar/www'
    assert server_root.line == 'ServerRoot "/home/skontar/www"'
    assert server_root.file_name == 'hello.conf'
    assert server_root.file_path == '/etc/httpd/conf.d/d1/hello.conf'


def test_empty_last_line():
    httpd = httpd_conf.HttpdConf(context_wrap(HTTPD_EMPTY_LAST, path='/etc/httpd/conf/httpd.conf'))
    result = HttpdConfTree([httpd])

    index_options = result['IndexOptions'][-1]
    assert index_options.value == 'FancyIndexing HTMLTable VersionSort'


def test_indented_lines_and_comments():
    httpd = httpd_conf.HttpdConf(context_wrap(HTTPD_EMBEDDED_QUOTES, path='/etc/httpd/conf/httpd.conf'))
    result = HttpdConfTree([httpd])

    request_headers = result['RequestHeader']
    assert len(request_headers) == 2


def test_regex_and_op_attrs():
    httpd = httpd_conf.HttpdConf(context_wrap(HTTPD_REGEX_AND_OP_ATTRS, path='/etc/httpd/conf/httpd.conf'))
    result = HttpdConfTree([httpd])

    rewrite_cond = result["RewriteCond"]
    assert len(rewrite_cond) == 3

    if_version = result["IfVersion"]
    assert len(if_version) == 1
    assert if_version.value == "< 2.4"


def test_unicode_comments():
    httpd = httpd_conf.HttpdConf(context_wrap(UNICODE_COMMENTS, path='/etc/httpd/conf/httpd.conf'))
    result = HttpdConfTree([httpd])

    rewrite_cond = result["DNSSDEnable"]
    assert len(rewrite_cond) == 1


def test_multiple_includes():
    httpd1 = httpd_conf.HttpdConf(context_wrap(MULTIPLE_INCLUDES, path='/etc/httpd/conf/httpd.conf'))
    httpd2 = httpd_conf.HttpdConf(context_wrap(UNICODE_COMMENTS, path='/etc/httpd/conf.d/05-foreman.d/hello.conf'))
    result = HttpdConfTree([httpd1, httpd2])
    assert len(result["IfVersion"]["DNSSDEnable"]) == 2


def test_recursive_includes():
    with pytest.raises(Exception):
        httpd1 = httpd_conf.HttpdConf(
            context_wrap(MULTIPLE_INCLUDES, path='/etc/httpd/conf/httpd.conf')
        )
        httpd2 = httpd_conf.HttpdConf(
            context_wrap(MULTIPLE_INCLUDES, path='/etc/httpd/conf.d/05-foreman.d/hello.conf')
        )
        HttpdConfTree([httpd1, httpd2])


def test_mixed_name():
    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_MIXED_NAME, path='/etc/httpd/conf/httpd.conf'))
    result = HttpdConfTree([httpd1])
    assert len(result.doc["H2Push"]) == 1


def test_empty_attr():
    httpd1 = httpd_conf.HttpdConf(context_wrap(HTTPD_EMPTY_ATTR, path='/etc/httpd/conf/httpd.conf'))
    result = HttpdConfTree([httpd1])
    assert len(result['VirtualHost']['RequestHeader']) == 4
    assert result['VirtualHost']['RequestHeader'][0].value == 'set X_FORWARDED_PROTO http'
    assert result['VirtualHost']['RequestHeader'][-1].value == 'set SSL_CLIENT_VERIFY ""'


def test_multiline_quote():
    with pytest.raises(ParseException):
        httpd_conf.HttpdConf(context_wrap(INVALID_CONF, path='/etc/httpd/conf/error.conf'))

    ml_quote = httpd_conf.HttpdConf(context_wrap(HTTPD_MULTILINE_QUOTE, path='/etc/httpd/conf.d/multiline-quote.conf'))
    httpd = httpd_conf.HttpdConf(context_wrap(HTTPD_CONF_1, path='/etc/httpd/conf/httpd.conf'))
    result = HttpdConfTree([httpd, ml_quote])

    assert result['JustFotTest_NoSec'].value == "/var/www/cgi"
    assert result['IfModule']['ServerLimit'].value == 256
    assert result['Listen'].value == 81

    # For the below tests, check the string value, then check the actual
    # attrs since quotes from quoted text are removed. The attrs store the
    # actual parsed attrs in a list form.
    assert result['VirtualHost']['SetEnv'].values == ['GIT_PROJECT_ROOT /var/www/git', 'GIT_HTTP_EXPORT_ALL']
    assert result['VirtualHost']['SetEnv'].children[0].attrs == ['GIT_PROJECT_ROOT', '/var/www/git']
    assert result['VirtualHost']['SetEnv'].children[1].attrs == ['GIT_HTTP_EXPORT_ALL']

    assert result['VirtualHost']['ProxyPass'].value == "http://backend.example.com/a-long-path-to-demonstrate connectiontimeout=1 retry=0"  # noqa
    assert result['VirtualHost']['ProxyPass'].children[-1].attrs == [
        'http://backend.example.com/a-long-path-to-demonstrate',
        'connectiontimeout=1',
        'retry=0'
    ]

    assert result['VirtualHost']['ScriptAliasMatch'].value == r"(?x)^/git/(.*/(HEAD | info/refs | objects/(info/[^/]+ | [0-9a-f]{2}/[0-9a-f]{38} | pack/pack-[0-9a-f]{40}\.(pack|idx)) | git-(upload|receive)-pack))$ /usr/libexec/git-core/git-http-backend/$1"  # noqa
    assert result['VirtualHost']['ScriptAliasMatch'].children[-1].attrs == [
        r'(?x)^/git/(.*/(HEAD | info/refs | objects/(info/[^/]+ | [0-9a-f]{2}/[0-9a-f]{38} | pack/pack-[0-9a-f]{40}\.(pack|idx)) | git-(upload|receive)-pack))$',  # noqa
        '/usr/libexec/git-core/git-http-backend/$1'
    ]



# coding: utf-8

from __future__ import print_function, absolute_import, division, unicode_literals

if False:  # MYPY
    from typing import Dict, Any  # NOQA

_package_data = dict(
    full_package_name='ruamel.yaml',
    version_info=(0, 16, 13),
    __version__='0.16.13',
    author='Anthon van der Neut',
    author_email='a.van.der.neut@ruamel.eu',
    description='ruamel.yaml is a YAML parser/emitter that supports roundtrip preservation of comments, seq/map flow style, and map key order',  # NOQA
    entry_points=None,
    since=2014,
    extras_require={
        ':platform_python_implementation=="CPython" and python_version<="2.7"': ['ruamel.ordereddict'],  # NOQA
        ':platform_python_implementation=="CPython" and python_version<"3.10"': ['ruamel.yaml.clib>=0.1.2'],  # NOQA
        'jinja2': ['ruamel.yaml.jinja2>=0.2'],
        'docs': ['ryd'],
    },
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: Implementation :: Jython',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup',
        'Typing :: Typed',
    ],
    keywords='yaml 1.2 parser round-trip preserve quotes order config',
    read_the_docs='yaml',
    supported=[(2, 7), (3, 5)],  # minimum
    tox=dict(
        env='*',  # remove 'pn', no longer test narrow Python 2.7 for unicode patterns and PyPy
        deps='ruamel.std.pathlib',
        fl8excl='_test/lib',
    ),
    universal=True,
    rtfd='yaml',
)  # type: Dict[Any, Any]


version_info = _package_data['version_info']
__version__ = _package_data['__version__']

try:
    from .cyaml import *  # NOQA

    __with_libyaml__ = True
except (ImportError, ValueError):  # for Jython
    __with_libyaml__ = False

from ...ruamel.yaml.main import *  # NOQA



"""
Custom datasource for gathering a list of the ethernet interface names.
"""
from insights.core.context import HostContext
from insights.core.dr import SkipComponent
from insights.core.plugins import datasource
from insights.core.spec_factory import simple_command
from insights.parsers.nmcli import NmcliConnShow
from insights.specs import Specs


class LocalSpecs(Specs):
    """ Local specs used only by ethernet_interfaces datasource. """
    ip_link = simple_command("/sbin/ip -o link")


@datasource(LocalSpecs.ip_link, HostContext)
def interfaces(broker):
    """
    This datasource provides a list of the ethernet interfaces available.

    Typical content of the spec is::

        1: lo: <LOOPBACK,UP,LOWER_UP> mtu 65536 qdisc noqueue state UNKNOWN mode DEFAULT group default qlen 1000\\    link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00
        2: enp1s0: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc fq_codel state UP mode DEFAULT group default qlen 1000\\    link/ether 52:54:00:13:14:b5 brd ff:ff:ff:ff:ff:ff
        3: enp8s0: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc fq_codel state UP mode DEFAULT group default qlen 1000\\    link/ether 52:54:00:e5:11:d4 brd ff:ff:ff:ff:ff:ff
        4: enp1s0.2@enp1s0: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc noqueue state UP mode DEFAULT group default qlen 1000\\    link/ether 52:54:00:13:14:b5 brd ff:ff:ff:ff:ff:ff
        5: ib0: <NO-CARRIER,BROADCAST,MULTICAST,UP> mtu 4092 qdisc mq state DOWN group default qlen 256\\    link/infiniband 00:01:02:03:fd:90:0:00:00:00:00:00:ef:0d:8b:02:01:d9:82:fd

    Note:
        This datasource may be executed using the following command:

        ``insights cat --no-header ethernet_interfaces``

    Sample data returned::

        ['enp1s0', 'enp8s0', 'enp1s0.2']

    Returns:
        list: List of the ethernet interfaces available.

    Raises:
        SkipComponent: When there is not any content.
    """
    content = broker[LocalSpecs.ip_link].content
    if content:
        ifaces = []
        for x in content:
            # Only process lines that have link/ether, this should exclude non ethernet devices.
            if "link/ether" in x:
                # Split first on : the interface name should be the second entry.
                # Then split again on @ since vlans append @ and the parent interface name to the end.
                ifaces.append(x.split(':')[1].split('@')[0].strip())

        if ifaces:
            return sorted(ifaces)

    raise SkipComponent


@datasource(NmcliConnShow, HostContext)
def team_interfaces(broker):
    """
    This datasource provides a list of the team device.

    Sample data returned::

        ['team0', 'team1']

    Returns:
        list: List of the team device.

    Raises:
        SkipComponent: When there is not any team interfaces.
    """

    content = broker[NmcliConnShow].data
    if content:
        team_ifaces = []
        for x in content:
            if 'team' in x['TYPE'] and x['DEVICE'] != '--':
                team_ifaces.append(x['DEVICE'])

        if team_ifaces:
            return sorted(team_ifaces)

    raise SkipComponent



"""
GaleraCnf - file ``/etc/my.cnf.d/galera.cnf``
=============================================
"""
from insights.core import IniConfigFile
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.galera_cnf)
class GaleraCnf(IniConfigFile):
    """
    This module provides parsing for the galera configuration of
    MySQL. The input is the contents of the file
    `/etc/my.cnf.d/galera.cnf`.

    Typical contents of the `galera.cnf` file looks like this::

        [client]
        port = 3306
        socket = /var/lib/mysql/mysql.sock

        [isamchk]
        key_buffer_size = 16M

        [mysqld]
        basedir = /usr
        binlog_format = ROW
        datadir = /var/lib/mysql
        default-storage-engine = innodb
        expire_logs_days = 10
        innodb_autoinc_lock_mode = 2
        innodb_locks_unsafe_for_binlog = 1
        key_buffer_size = 16M
        log-error = /var/log/mariadb/mariadb.log
        max_allowed_packet = 16M
        max_binlog_size = 100M
        max_connections = 8192
        wsrep_max_ws_rows = 131072
        wsrep_max_ws_size = 1073741824

        [mysqld_safe]
        log-error = /var/log/mariadb/mariadb.log
        nice = 0
        socket = /var/lib/mysql/mysql.sock

        [mysqldump]
        max_allowed_packet = 16M
        quick
        quote-names

    Examples:
        >>> type(galera_conf)
        <class 'insights.parsers.galera_cnf.GaleraCnf'>
        >>> 'mysqld' in galera_conf
        True
        >>> 'client' in galera_conf
        True
        >>> galera_conf.has_option('isamchk', 'key_buffer_size')
        True
        >>> galera_conf.has_option('mysqld', 'foo')
        False
        >>> galera_conf.get('client', 'port')
        '3306'
        >>> expected = {'port': '3306', 'socket': '/var/lib/mysql/mysql.sock'}
        >>> galera_conf.items('client') == expected
        True
    """
    def parse_content(self, content, allow_no_value=True):
        super(GaleraCnf, self).parse_content(content, allow_no_value)



import logging
import os
import six
from contextlib import contextmanager
from insights.util import streams, subproc

log = logging.getLogger(__name__)
GLOBAL_PRODUCTS = []
PRODUCT_NAMES = []
DEFAULT_VERSION = ["-1", "-1"]

FSRoots = []


def fs_root(thing):
    FSRoots.append(thing)
    return thing


def product(klass):
    GLOBAL_PRODUCTS.append(klass)
    PRODUCT_NAMES.append(klass.name)
    return klass


def get_system(metadata, hostname):
    for system in metadata.get("systems", []):
        if system.get("system_id") == hostname:
            return system


class MultiNodeProduct(object):

    def __init__(self, role=None):
        self.role = role

    def __bool__(self):
        return bool(self.role)

    __nonzero__ = __bool__

    def is_parent(self):
        return self.role == self.parent_type


def create_product(metadata, hostname):
    current_system = get_system(metadata, hostname)
    for p in GLOBAL_PRODUCTS:
        if metadata.get("product", "").lower() == p.name and current_system:
            instance = p()
            instance.__dict__ = current_system
            if hasattr(instance, "type"):
                instance.role = instance.type
            return instance


@product
class Docker(MultiNodeProduct):

    name = "docker"
    parent_type = "host"


@product
class OSP(MultiNodeProduct):

    name = "osp"
    parent_type = "Director"


@product
class RHEV(MultiNodeProduct):

    name = "rhev"
    parent_type = "Manager"


@product
class RHEL(object):

    name = "rhel"

    def __init__(self, version=DEFAULT_VERSION, release=None):
        self.version = version
        self.release = release

    def __bool__(self):
        return all([(self.version != DEFAULT_VERSION),
                   bool(self.release)])

    __nonzero__ = __bool__

    @classmethod
    def from_metadata(cls, metadata, processor_obj):
        return cls()


class Context(object):
    def __init__(self, **kwargs):
        self.version = kwargs.pop("version", DEFAULT_VERSION)
        self.metadata = kwargs.pop("metadata", {})
        self.loaded = True
        self.cmd = None
        optional_attrs = [
            "content", "path", "hostname", "release",
            "machine_id", "target", "last_client_run", "relative_path",
            "args"
        ]
        for k in optional_attrs:
            setattr(self, k, kwargs.pop(k, None))

        self.relative_path = self.path

        for p in GLOBAL_PRODUCTS:
            if p.name in kwargs:
                setattr(self, p.name, kwargs.pop(p.name))
            else:
                setattr(self, p.name, create_product(self.metadata, self.hostname))

    def stream(self):
        return iter(self.content)

    def product(self):
        for pname in PRODUCT_NAMES:
            if hasattr(self, pname):
                return getattr(self, pname)

    def __repr__(self):
        return repr(dict((k, str(v)[:30]) for k, v in self.__dict__.items()))


class ExecutionContextMeta(type):
    registry = []

    def __init__(cls, name, bases, dct):
        if name == "ExecutionContext":
            return
        ExecutionContextMeta.registry.append(cls)

    # Remember that contexts are tried *in reverse order* so that they
    # may be overridden by just loading a plugin.
    @classmethod
    def identify(cls, files):
        for e in reversed(cls.registry):
            root, ctx = e.handles(files)
            if ctx is not None:
                return (root, ctx)
        return (None, None)


class ExecutionContext(six.with_metaclass(ExecutionContextMeta)):
    marker = None

    def __init__(self, root="/", timeout=None, all_files=None):
        self.root = root
        self.timeout = timeout
        self.all_files = all_files or []

    @classmethod
    def handles(cls, files):
        if cls.marker is None or not files:
            return (None, None)

        sep = os.path.sep
        m = sep + cls.marker.lstrip(sep)
        marker_root = set()
        for f in files:
            if m in f:
                i = f.find(m)
                if f.endswith(m) or f[i + len(m)] == sep:
                    root = os.path.dirname(f[:i + 1])
                    marker_root.add(root)
        if len(marker_root) == 1:
            return (marker_root.pop(), cls)
        if len(marker_root) > 1:
            # when more marker found, return the one which is closest to root
            closest_root = marker_root.pop()
            for left_one in marker_root:
                if len(left_one) < len(closest_root):
                    closest_root = left_one
            return (closest_root, cls)
        return (None, None)

    def check_output(self, cmd, timeout=None, keep_rc=False, env=None, signum=None):
        """ Subclasses can override to provide special
            environment setup, command prefixes, etc.
        """
        return subproc.call(cmd, timeout=timeout or self.timeout, signum=signum,
                keep_rc=keep_rc, env=env)

    def shell_out(self, cmd, split=True, timeout=None, keep_rc=False, env=None, signum=None):
        env = env or os.environ
        rc = None
        raw = self.check_output(cmd, timeout=timeout, keep_rc=keep_rc, env=env, signum=signum)
        if keep_rc:
            rc, output = raw
        else:
            output = raw

        if split:
            output = output.splitlines()

        return (rc, output) if keep_rc else output

    @contextmanager
    def stream(self, *args, **kwargs):
        with streams.stream(*args, **kwargs) as s:
            yield s

    @contextmanager
    def connect(self, *args, **kwargs):
        with streams.connect(*args, **kwargs) as s:
            yield s

    def locate_path(self, path):
        return os.path.expandvars(path)

    def __repr__(self):
        msg = "<%s('%s', %s)>"
        return msg % (self.__class__.__name__, self.root, self.timeout)


@fs_root
class HostContext(ExecutionContext):
    def __init__(self, root='/', timeout=30, all_files=None):
        super(HostContext, self).__init__(root=root, timeout=timeout, all_files=all_files)


@fs_root
class HostArchiveContext(ExecutionContext):
    marker = "insights_commands"


@fs_root
class SerializedArchiveContext(ExecutionContext):
    marker = "insights_archive.txt"


@fs_root
class SosArchiveContext(ExecutionContext):
    marker = "sos_commands"


@fs_root
class ClusterArchiveContext(ExecutionContext):
    pass


@fs_root
class DockerImageContext(ExecutionContext):
    pass


@fs_root
class JBossContext(HostContext):
    pass


@fs_root
class JDRContext(ExecutionContext):
    marker = "JBOSS_HOME"

    def locate_path(self, path):
        p = path.replace("$JBOSS_HOME", "JBOSS_HOME")
        return super(JDRContext, self).locate_path(p)


@fs_root
class InsightsOperatorContext(ExecutionContext):
    """Recognizes insights-operator archives"""
    marker = "config/featuregate"


@fs_root
class MustGatherContext(ExecutionContext):
    """Recognizes must-gather archives"""
    marker = "cluster-scoped-resources"


class OpenStackContext(ExecutionContext):
    def __init__(self, hostname):
        super(OpenStackContext, self).__init__()
        self.hostname = hostname



from insights.parsers import ktimer_lockless
from insights.parsers.ktimer_lockless import KTimerLockless
from insights.parsers import SkipException
from insights.tests import context_wrap
import pytest
import doctest

KTIMER_LOCKLESS_OUTPUT = "0"

KTIMER_LOCKLESS_EMPTY = ""


def test_ktimer_lockless_parser():
    ktimer_lockless_obj = ktimer_lockless.KTimerLockless(context_wrap(KTIMER_LOCKLESS_OUTPUT))
    assert ktimer_lockless_obj.ktimer_lockless_val == 0


def test_empty():
    with pytest.raises(SkipException) as e:
        KTimerLockless(context_wrap(KTIMER_LOCKLESS_EMPTY))
    assert 'The file is empty' in str(e)


def test_ktimer_lockless_examples():
    env = {
        'ktimer_lockless': KTimerLockless(context_wrap(KTIMER_LOCKLESS_OUTPUT)),
    }
    failed, total = doctest.testmod(ktimer_lockless, globs=env)
    assert failed == 0



"""
PacemakerLog - file ``/var/log/pacemaker.log``
==============================================
"""

from .. import LogFileOutput, parser
from insights.specs import Specs


@parser(Specs.pacemaker_log)
class PacemakerLog(LogFileOutput):
    """
    Read the pacemaker log file.  Uses the ``LogFileOutput`` class parser
    functionality.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`


    Sample pacemaker.log::

        Aug 21 12:58:40 [11656] example.redhat.com        cib:     info: crm_client_destroy: 	Destroying 0 events
        Aug 21 12:59:53 [11655] example.redhat.com pacemakerd:     info: pcmk_quorum_notification: 	Membership 12: quorum retained (3)
        Aug 21 12:59:53 [11661] example.redhat.com       crmd:     info: pcmk_quorum_notification: 	Membership 12: quorum retained (3)
        Aug 21 12:59:53 [11655] example.redhat.com pacemakerd:     info: pcmk_quorum_notification: 	Membership 12: quorum retained (3)

    .. note::
        Because pacemaker timestamps by default have no year, the
        year of the logs will be inferred from the year in your timestamp.
        This will also work around December/January crossovers.

    Examples:
        >>> pm = shared(PacemakerLog)
        >>> pm.get('crmd')[0]['raw_message']
        'Aug 21 12:59:53 [11661] example.redhat.comm       crmd:     info: pcmk_quorum_notification: 	Membership 12: quorum retained (3)'
        >>> from datetime import datetime
        >>> len(list(pm.get_after(datetime(21, 8, 2017, 12, 59, 50))))
        3
    """
    time_format = '%b %d %H:%M:%S'



from insights.parsers.proc_limits import HttpdLimits, MysqldLimits, OvsVswitchdLimits
from insights.tests import context_wrap
from insights.parsers import ParseException
import pytest

PROC_LIMITS_ERR = """
cat: /proc/1100/limits: No such file or directory
""".strip()

PROC_LIMITS = """
Limit                     Soft Limit           Hard Limit           Units
Max cpu time              unlimited            unlimited            seconds
Max file size             unlimited            unlimited            bytes
Max data size             unlimited            unlimited            bytes
Max stack size            10485760             unlimited            bytes
Max core file size        0                    unlimited            bytes
Max resident set          unlimited            unlimited            bytes
Max processes             9                    99                   processes
Max open files            1024                 4096                 files
Max locked memory         65536                65536                bytes
Max address space         unlimited            unlimited            bytes
Max file locks            unlimited            unlimited            locks
Max pending signals       15211                15211                signals
Max msgqueue size         819200               819200               bytes
Max nice priority         0                    0
Max realtime priority     0                    0
Max realtime timeout      unlimited            unlimited            us
""".strip()


def test_httpd_limits():
    results = HttpdLimits(context_wrap(PROC_LIMITS))
    assert len(results) == 16
    assert "max_processes" in results
    assert results.max_processes.hard_limit == '99'
    assert results.max_processes.soft_limit == '9'
    assert results.max_processes.units == 'processes'
    assert results.max_nice_priority.hard_limit == '0'
    assert results.max_nice_priority.units == ''
    for r in results:
        if 'Max cpu time' == r['Limit']:
            assert r['Hard_Limit'] == 'unlimited'
            assert r['Soft_Limit'] == 'unlimited'
        if 'Max realtime timeout' == r['Limit']:
            assert r['Units'] == 'us'


def test_proc_limits_err():
    with pytest.raises(ParseException) as pe:
        HttpdLimits(context_wrap(PROC_LIMITS_ERR))
        assert PROC_LIMITS_ERR in str(pe)

    with pytest.raises(ParseException) as pe:
        HttpdLimits(context_wrap(''))
        assert 'empty file' in str(pe)


def test_mysqld_limits():
    results = MysqldLimits(context_wrap(PROC_LIMITS))
    assert results.max_open_files.hard_limit == '4096'
    assert results.max_open_files.soft_limit == '1024'
    assert len(results) == 16


def test_ovs_vswitchd_limits():
    results = OvsVswitchdLimits(context_wrap(PROC_LIMITS))
    assert "max_open_files" in results
    assert results.max_open_files.units == 'files'
    assert results.max_open_files.hard_limit == '4096'



"""
passenger-status command
========================
This module provides processing for the ``passenger-status`` command using the
following parsers:
"""
from insights.parsers import SkipException
from insights.specs import Specs
from insights import CommandParser, parser


@parser(Specs.passenger_status)
class PassengerStatus(CommandParser, dict):
    """
    Parse the passenger-status command output.

    Produces a simple dictionary of keys and values from the command output
    contents.

    Sample command output::

        Version : 4.0.18
        Date    : 2018-10-23 15:42:04 +0800
        Instance: 1265
        ----------- General information -----------
        Max pool size : 12
        Processes     : 2
        Requests in top-level queue : 0

        ----------- Application groups -----------
        /usr/share/foreman#default:
        App root: /usr/share/foreman
        Requests in queue: 192
        * PID: 30131   Sessions: 1       Processed: 991     Uptime: 2h 9m 8s
        CPU: 3%      Memory  : 562M    Last used: 1h 53m 51s
        * PID: 32450   Sessions: 1       Processed: 966     Uptime: 2h 8m 15s
        CPU: 4%      Memory  : 463M    Last used: 1h 48m 17
        * PID: 4693    Sessions: 1       Processed: 939     Uptime: 2h 6m 32s
        CPU: 3%      Memory  : 470M    Last used: 1h 50m 48

        /etc/puppet/rack#default:
        App root: /etc/puppet/rack
        Requests in queue: 0
        * PID: 21934   Sessions: 1       Processed: 380     Uptime: 1h 33m 34s
        CPU: 1%      Memory  : 528M    Last used: 1h 29m 4
        * PID: 26194   Sessions: 1       Processed: 544     Uptime: 1h 31m 34s
        CPU: 2%      Memory  : 490M    Last used: 1h 23m 5
        * PID: 32384   Sessions: 1       Processed: 36      Uptime: 1h 0m 29s
        CPU: 0%      Memory  : 561M    Last used: 1h 0m 3s

    Examples:
        >>> passenger_status["Version"]
        '4.0.18'
        >>> 'rack_default' in passenger_status
        True
        >>> len(passenger_status['foreman_default']['p_list'])
        3

    Raises:
        SkipException: When input content is empty or there is no useful data.
    """

    def parse_content(self, content):
        if len(content) <= 1:
            raise SkipException("Empty content")

        group = ''
        data = {}
        for line in content:
            line = line.strip()
            if not group and ":" in line and not line.endswith(':'):
                key, val = [i.strip() for i in line.split(':', 1)]
                data[key] = val
            elif line.startswith(("/usr/share/foreman#default", "/etc/puppet/rack#default")):
                # set the group which the following pid will belong to
                group = line.strip(':').split('/')[-1].replace('#', '_')
                data[group] = group_list = {}
            elif group:
                if line.startswith(("* PID:", "CPU")):
                    if line.strip().startswith("* PID:"):
                        p_dict = {}
                        if 'p_list' not in group_list:
                            group_list['p_list'] = []
                        group_list['p_list'].append(p_dict)
                    l = line.lstrip('* ')
                    # parse each key: value pair
                    while (':' in l and l.index(':') != l.rindex(':')):
                        k, l = [i.strip() for i in l.split(':', 1)]
                        v, l = [i.strip() for i in l.split(None, 1)]
                        p_dict[k] = v
                    # for the last key: value pair
                    if l and ':' in l:
                        k, v = [i.strip() for i in l.split(':')]
                        p_dict[k] = v
                elif ':' in line:
                    key, val = [i.strip() for i in line.split(':', 1)]
                    group_list[key] = val

        if not data:
            raise SkipException("No useful data")
        self.update(data)

    @property
    def data(self):
        """
        (dict): A simple dictionary of keys and values from the command output contents
        """
        return self



import pytest
import doctest
from insights.parsers import SkipException, losetup
from insights.parsers.losetup import LoSetup
from insights.tests import context_wrap

LOSETUP_RHEL8 = """
NAME       SIZELIMIT OFFSET AUTOCLEAR RO BACK-FILE       DIO LOG-SEC
/dev/loop2      4096    256         0  1 /root/disk2.img   0    1024
/dev/loop0         0      0         1  0 /root/disk1.img   1     512
"""

LOSETUP_RHEL7 = """
NAME       SIZELIMIT OFFSET AUTOCLEAR RO BACK-FILE
/dev/loop3      4096    256         0  1 /root/disk3.img
"""

LOSETUP_EMPTY = ""

LOSETUP_DOCTEST = """
NAME       SIZELIMIT OFFSET AUTOCLEAR RO BACK-FILE      DIO LOG-SEC
/dev/loop0         0      0         0  0 /root/disk.img   1     512
"""


def test_losetup():
    losetup = LoSetup(context_wrap(LOSETUP_RHEL8))
    assert len(losetup) == 2
    assert losetup[0]['NAME'] == '/dev/loop2'
    assert losetup[0]['SIZELIMIT'] == 4096
    assert losetup[0]['OFFSET'] == 256
    assert losetup[0]['AUTOCLEAR'] is False
    assert losetup[0]['RO'] is True
    assert losetup[0]['BACK-FILE'] == '/root/disk2.img'
    assert losetup[0]['DIO'] is False
    assert losetup[0]['LOG-SEC'] == 1024
    assert losetup[1]['NAME'] == '/dev/loop0'
    assert losetup[1]['SIZELIMIT'] == 0
    assert losetup[1]['OFFSET'] == 0
    assert losetup[1]['AUTOCLEAR'] is True
    assert losetup[1]['RO'] is False
    assert losetup[1]['BACK-FILE'] == '/root/disk1.img'
    assert losetup[1]['DIO'] is True
    assert losetup[1]['LOG-SEC'] == 512

    losetup = LoSetup(context_wrap(LOSETUP_RHEL7))
    assert len(losetup) == 1
    assert losetup[0]['NAME'] == '/dev/loop3'
    assert losetup[0]['SIZELIMIT'] == 4096
    assert losetup[0]['OFFSET'] == 256
    assert losetup[0]['AUTOCLEAR'] is False
    assert losetup[0]['RO'] is True
    assert losetup[0]['BACK-FILE'] == '/root/disk3.img'
    assert 'DIO' not in losetup[0]
    assert 'LOG-SEC' not in losetup[0]

    with pytest.raises(SkipException):
        LoSetup(context_wrap(LOSETUP_EMPTY))


def test_losetup_doc_examples():
    env = {'losetup': LoSetup(context_wrap(LOSETUP_DOCTEST))}
    failed, total = doctest.testmod(losetup, globs=env)
    assert failed == 0



"""
BDIReadAheadKB - file ``/sys/class/bdi/*/read_ahead_kb``
========================================================

This parser reads the content of ``/sys/class/bdi/*/read_ahead_kb``.
"""

from insights import Parser, parser
from insights.specs import Specs
from insights.parsers import ParseException


@parser(Specs.bdi_read_ahead_kb)
class BDIReadAheadKB(Parser):
    """
    Class ``BDIReadAheadKB`` parses the content of the ``/sys/class/bdi/*/read_ahead_kb``.

    Raises:
        ParseException: When content is empty or unparseable

    A typical sample of the content of this file looks like::

        128

    Examples:
        >>> type(bdi_read_ahead_kb)
        <class 'insights.parsers.bdi_read_ahead_kb.BDIReadAheadKB'>
        >>> bdi_read_ahead_kb.read_ahead_kb
        128
    """

    def parse_content(self, content):
        if len(content) != 1:
            raise ParseException('Error: {0}'.format(content if content else 'empty file'))
        try:
            self._read_ahead_kb = int(content[0].strip())
        except ValueError:
            raise ParseException("Error: unparseable content: {0}".format(content[0]))

    @property
    def read_ahead_kb(self):
        """ int: Value of ``read_ahead_kb``."""
        return self._read_ahead_kb



import json

from insights.parsr.query import from_dict, make_child_query as q


DATA = json.loads('''
{
  "kind": "ClusterVersion",
  "apiVersion": "config.openshift.io/v1",
  "metadata": {
    "name": "version",
    "selfLink": "/apis/config.openshift.io/v1/clusterversions/version",
    "uid": "11111111-2222-3333-4444-555555555555",
    "resourceVersion": "1",
    "generation": 1,
    "creationTimestamp": "2019-08-04T23:16:46Z"
  },
  "spec": {
    "clusterID": "55555555-4444-3333-2222-111111111111",
    "upstream": "xxxxx://xxx.xxxxxxxxx.xxx/xxx/xxxxxxxxxxxxx/xx/xxxxx",
    "channel": "stable-4.2"
  },
  "status": {
    "desired": {
      "version": "4.2.0-0.ci-2019-08-04-183142",
      "image": "registry.svc.ci.openshift.org/ocp/release@sha256:63b65452005d6e9e45bb92a7505524db0e406c3281d91bdc1a4f5c5cf71b01c5",
      "force": false
    },
    "history": [
      {
        "state": "Completed",
        "startedTime": "2019-08-04T23:17:08Z",
        "completionTime": "2019-08-04T23:32:14Z",
        "version": "4.2.0-0.ci-2019-08-04-183142",
        "image": "registry.svc.ci.openshift.org/ocp/release@sha256:63b65452005d6e9e45bb92a7505524db0e406c3281d91bdc1a4f5c5cf71b01c5",
        "verified": false
      }
    ],
    "observedGeneration": 1,
    "versionHash": "############",
    "conditions": [
      {
        "type": "Available",
        "status": "True",
        "lastTransitionTime": "2019-08-04T23:32:14Z",
        "message": "Done applying 4.2.0-0.ci-2019-08-04-183142"
      },
      {
        "type": "Failing",
        "status": "True",
        "lastTransitionTime": "2019-08-05T15:04:39Z",
        "reason": "ClusterOperatorNotAvailable",
        "message": "Cluster operator console is still updating"
      },
      {
        "type": "Progressing",
        "status": "False",
        "lastTransitionTime": "2019-08-04T23:32:14Z",
        "reason": "ClusterOperatorNotAvailable",
        "message": "Error while reconciling 4.2.0-0.ci-2019-08-04-183142: the cluster operator console has not yet successfully rolled out"
      },
      {
        "type": "RetrievedUpdates",
        "status": "False",
        "lastTransitionTime": "2019-08-04T23:17:08Z",
        "reason": "RemoteFailed",
        "message": "Unable to retrieve available updates: currently installed version 4.2.0-0.ci-2019-08-04-183142 not found in the stable-4.2 channel"
      }
    ],
    "availableUpdates": null
  }
}
''')

d = from_dict(DATA)


def test_values():
    assert d.status.conditions.status.values == ["True", "True", "False", "False"]


def test_where():
    assert len(d.status.conditions.where("type", "Progressing")) == 1

    res = d.status.conditions.where(q("type", "Progressing") | q("status", "True"))
    assert len(res) == 3

    # because only two of the resulting conditions have children named reason.
    assert res.reason.values == ["ClusterOperatorNotAvailable", "ClusterOperatorNotAvailable"]


def test_where_lambda():
    assert bool(d.status.conditions.where(lambda n: n.type.value == "Progressing"))
    assert bool(d.status.where(lambda n: len(n.conditions) > 3))
    assert not bool(d.status.where(lambda n: len(n.conditions) > 9))



"""
System time configuration
=========================

This is a collection of parsers that all deal with the system's configuration
of its time setting resources.

"""

import re
from .. import Parser, parser, get_active_lines, CommandParser
from insights.specs import Specs


class NTPConfParser(Parser):
    """
    NTP and Chrony both use the same format for their configuration file -
    a series of keywords with optional values.  Some keywords can appear
    more than once, so all keyword values are stored as a list of strings.
    Keywords that have no value, like 'iburst' or 'rtcsync', are left as
    keys but have None as a value.

    Also provides the ``servers`` and ``peers`` properties as (sorted) lists
    of the found 'server' and 'peer' data (respectively).

    Sample Input::
        >>> ntp_conf_data = '''
        ... server 0.rhel.pool.ntp.org iburst
        ... server 1.rhel.pool.ntp.org iburst
        ... server 2.rhel.pool.ntp.org iburst
        ... server 3.rhel.pool.ntp.org iburst
        ... # Enable kernel RTC synchronization.
        ... rtcsync
        ... leapsecmode slew
        ... maxslewrate 1000
        ... smoothtime 400 0.001 leaponly
        ... tinker step 0.9
        ... '''

    Examples:

        >>> ntp = shared[NTP_conf]
        >>> 'rtcsync' in ntp.data # Single word options are present but None
        True
        >>> ntp.data['rtcsync'] # Not in dictionary if option not set
        None
        >>> len(ntp.data['server'])
        4
        >>> ntp.data['server'][0]
        '0.rhel.pool.ntp.org iburst'
        >>> ntp.servers[0] # same data as above
        '0.rhel.pool.ntp.org iburst'
        >>> ntp.data['maxslewrate']
        '1000'
        >>> ntp.get_last('rtcsync') # See above for fetching single-word options
        None
        >>> ntp.get_last('leapsecmode')
        'slew'
        >>> ntp.get_last('tinker', 'panic', 'none') # Use default value
        'none'
        >>> ntp.get_last('tinker', 'step', '1') # Use value set in file
        '0.9'
        >>> ntp.get_param('tinker', 'step') # Get list of all settings
        ['0.9']
    """
    def parse_content(self, content):
        data = {}
        for line in get_active_lines(content):
            if ' ' in line or '\t' in line:
                k, rest = line.split(None, 1)
                if k in data:
                    data[k].append(rest)
                else:
                    data[k] = [rest]
            else:
                data[line] = None
        self.data = data

        # Also set up some convenience access to lists of stuff:
        if 'server' in data:
            self.servers = sorted(data['server'])
        else:
            self.servers = []
        if 'peer' in data:
            self.peers = sorted(data['peer'])
        else:
            self.peers = []

    def get_param(self, keyword, param=None, default=None):
        """
        Get all the parameters for a given keyword, or default if keyword or
        parameter are not present in the configuration.

        This finds every declaration of the given parameter (which is the
        one which takes effect).  If no parameter is given, then the entire
        line is treated as the parameter.  There is always at least one
        element returned - the default, or

        Parameters:
            keyword(str): The keyword name, e.g. 'tinker' or 'driftfile'
            param(str): The parameter name, e.g. 'panic' or 'step'.  If not
                given, all the definitions of that keyword are given.
            default(str): The default (singular) value if the keyword or
                parameter is not found.  If not given, None is used.

        Returns:
            list: All the values of the given parameter, or an empty list if
            not found.
        """
        if not keyword or keyword not in self.data:
            return [default]
        # keyword in data - if no value, we store None, so return that in a list
        if self.data[keyword] is None:
            return [None]
        # If we're not searching for a particular parameter, just return all
        # the values for this keyword.
        if not param:
            return self.data[keyword]

        found = []
        for line in self.data[keyword]:
            # Line has already had keyword removed.
            words = line.strip().split()
            if len(words) > 1:
                # Line has param and value - check param:
                if words[0] == param:
                    found.append(words[1])
            else:
                found.append(words[0])
        if found == []:
            return [default]
        else:
            return found

    def get_last(self, keyword, param=None, default=None):
        """
        Get the parameters for a given keyword, or default if keyword or
        parameter are not present in the configuration.

        This finds the last declaration of the given parameter (which is the
        one which takes effect).  If no parameter is given, then the entire
        line is treated as the parameter and returned.

        Parameters:
            keyword(str): The keyword name, e.g. 'tinker' or 'driftfile'
            param(str): The parameter name, e.g. 'panic' or 'step'.  If not
                given, the last definition of that keyword is given.

        Returns:
            str or None: The value of the given parameter, or None if not
            found.
        """
        return self.get_param(keyword, param, default)[-1]


@parser(Specs.chrony_conf)
class ChronyConf(NTPConfParser):
    """
    A parser for analyzing the chrony service config file ``/etc/chrony.conf``

    Uses the ``NTPConfParser`` class defined in this module.
    """
    pass


@parser(Specs.ntp_conf)
class NTPConf(NTPConfParser):
    """
    A parser for analyzing the ntpd service config file ``/etc/ntp.conf``

    Uses the ``NTPConfParser`` class defined in this module.
    """
    pass


@parser(Specs.localtime)
class LocalTime(CommandParser):
    """
    A parser for working with the output of command: `file -L /etc/localtime`

    Sample Input::

        /etc/localtime: timezone data, version 2, 5 gmt time flags, 5 std time flags, no leap seconds, 69 transition times, 5 abbreviation chars

    Examples:

        >>> localtime = shared[LocalTime]
        >>> localtime.data['name']
        '/etc/localtime'
        >>> localtime.data['version']
        '2'
        >>> localtime.data['gmt_time_flag']
        '5'
        >>> localtime.data['leap_second']
        'no'
    """
    def parse_content(self, content):
        result = {}
        title = [
            'version',
            'gmt_time_flag',
            'std_time_flag',
            'leap_second',
            'transition_time',
            'abbreviation_char'
        ]
        for line in content:
            filename, info = line.strip().split(':', 1)
            result['name'] = filename.strip()
            info_list = info.split(', ')
            if len(info_list) == 7 and info_list[0].strip() == 'timezone data':
                # A bit of a hack - version field has the data second, all
                # other fields have the data first.
                for idx, k in enumerate(title):
                    if idx == 0:
                        result[k] = info_list[idx + 1].strip().split()[1]
                    else:
                        result[k] = info_list[idx + 1].strip().split()[0]
        self.data = result


@parser(Specs.ntptime)
class NtpTime(CommandParser):
    """
    A parser for working with the output of the ``ntptime``.

    This doesn't attempt to get much out of the output; useful things that
    it retrieves are:

    * ``ntp_gettime`` - the return code of the ``ntp_gettime()`` call.
    * ``ntp_adjtime`` - the return code of the ``ntp_adjtime()`` call.
    * ``status`` - the hexadecimal status code as a string.
    * ``flags`` - the flags in brackets after the status code.

    Sample Input::

        ntp_gettime() returns code 0 (OK)
          time dbbc595d.1adbd720  Thu, Oct 27 2016 18:45:49.104, (.104917550),
          maximum error 263240 us, estimated error 102 us, TAI offset 0
        ntp_adjtime() returns code 0 (OK)
          modes 0x0 (),
          offset 0.000 us, frequency 4.201 ppm, interval 1 s,
          maximum error 263240 us, estimated error 102 us,
          status 0x2011 (PLL,INS,NANO),
          time constant 2, precision 0.001 us, tolerance 500 ppm,

    Examples:

        >>> ntptime = shared[NtpTime]
        >>> ntptime.data['stats']
        '0x2011'
        >>> ntptime.data['ntp_gettime']
        '0'
        >>> ntptime.data['flags']
        ['PLL', 'INS', 'NANO']
        >>> ntptime.data['interval']  # Other values are integers or floats
        1
        >>> ntptime.data['precision']  # Note floats may not be exact
        0.001
        >>> ntptime.data['maximum error']
        263240

    """
    def parse_content(self, content):
        result = {}
        return_code_re = re.compile(r'(?P<func>ntp_(?:get|adj)time)\(\) returns code (?P<code>\d+) ')
        status_code_re = re.compile(r'status (?P<status>0x[0-9a-f]+) \((?P<flags>.*)\)')
        time_data_re = re.compile(r'time (?P<timecode>[0-9a-f]{8}\.[0-9a-f]{8}) +(?P<timestamp>\w{3}, \w{3} \d+ .*),')
        value_re = re.compile(r'(?P<keyword>[A-Za-z][A-Za-z ]+) (?P<value>\d+(?:\.\d+|x[0-9a-f]+)?) ?(?P<units>us|ppm|s|\(.*\))?')
        # Note that maximum and estimated error appear in both sections but
        # have the same value, so we don't bother to make them unique
        for line in content:
            # Check for function line
            match = return_code_re.search(line)
            if match:
                function = match.group('func')
                result[function] = match.group('code')
                continue

            # Check for status line specially
            match = status_code_re.search(line)
            if match:
                result['status'] = match.group('status')
                result['flags'] = match.group('flags').split(',')
                continue

            # Check for time data specially
            match = time_data_re.search(line)
            if match:
                # Get timecode and timestamp in one go
                result.update(match.groupdict())
                continue

            # Otherwise, try to read comma-separated key-value groups
            for match in value_re.finditer(line):
                value = match.group('value')
                if '.' in value:
                    vnum = float(value)
                elif 'x' in value:
                    vnum = int(value, 16)
                else:
                    vnum = int(value)
                result[match.group('keyword')] = vnum

        self.data = result



"""
LoSetup - command ``/usr/sbin/losetup -l``
==========================================

This parser reads the output of ``/usr/sbin/losetup -l`` into a list of entries.
Each entry is a dictionary of headers:

* ``NAME`` - the path name of the loop back device (strings)
* ``SIZELIMIT`` - the data end position of backing file in bytes (integer)
* ``OFFSET`` - the data start position of backing file in bytes (integer)
* ``AUTOCLEAR`` - the autoclear flag (boolean)
* ``RO`` - the read only flag (boolean)
* ``BACK-FILE`` - the path of the backing file (strings)
* ``DIO`` - the direct I/O flag (boolean)
* ``LOG-SEC`` - the logical sector size of the loop device in bytes (integer)

Sample output of ``losetup -l`` command is::

    NAME       SIZELIMIT OFFSET AUTOCLEAR RO BACK-FILE      DIO LOG-SEC
    /dev/loop0         0      0         0  0 /root/disk.img   1     512

Examples:

    >>> type(losetup)
    <class 'insights.parsers.losetup.LoSetup'>
    >>> len(losetup)
    1
    >>> losetup[0]['NAME']
    '/dev/loop0'
    >>> losetup[0]['RO']
    False
    >>> losetup[0]['DIO']
    True
    >>> losetup[0]['NAME']
    '/dev/loop0'
    >>> losetup[0]['SIZELIMIT']
    0
    >>> losetup[0]['OFFSET']
    0
    >>> losetup[0]['AUTOCLEAR']
    False
    >>> losetup[0]['RO']
    False
    >>> losetup[0]['BACK-FILE']
    '/root/disk.img'
    >>> losetup [0]['DIO']
    True
    >>> losetup[0]['LOG-SEC']
    512
"""
from insights import CommandParser, parser
from insights.specs import Specs
from insights.parsers import parse_delimited_table, SkipException


@parser(Specs.losetup)
class LoSetup(CommandParser, list):
    """
    Parses the output of the ``/usr/sbin/losetup -l`` command.
    """

    def parse_content(self, content):
        if not content:
            raise SkipException("Empty output.")

        self.extend(parse_delimited_table(content))
        for entry in self:
            for key in ['SIZELIMIT', 'OFFSET', 'LOG-SEC']:
                if key in entry:
                    entry[key] = int(entry[key])
            for key in ['AUTOCLEAR', 'RO', 'DIO']:
                if key in entry:
                    entry[key] = True if entry[key] == '1' else False



"""
Cinder configuration - file ``/etc/cinder/cinder.conf``
=======================================================
"""
from insights.core import IniConfigFile
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.cinder_conf)
class CinderConf(IniConfigFile):
    """
    The Cinder configuration file is a standard '.ini' file and this parser uses
    the ``IniConfigFile`` class to read it.

    Sample configuration::

        [DEFAULT]
        rpc_backend=cinder.openstack.common.rpc.impl_kombu
        control_exchange=openstack
        osapi_volume_listen=10.22.100.58
        osapi_volume_workers=32
        api_paste_config=/etc/cinder/api-paste.ini
        glance_api_servers=http://10.22.120.50:9292
        glance_api_version=2
        glance_num_retries=0
        glance_api_insecure=False
        glance_api_ssl_compression=False
        enable_v1_api=True
        enable_v2_api=True
        storage_availability_zone=nova
        default_availability_zone=nova
        enabled_backends=tripleo_ceph
        nova_catalog_info=compute:Compute Service:publicURL
        nova_catalog_admin_info=compute:Compute Service:adminURL

        [lvm]
        iscsi_helper=lioadm
        volume_group=cinder-volumes
        iscsi_ip_address=192.168.88.10
        volume_driver=cinder.volume.drivers.lvm.LVMVolumeDriver
        volumes_dir=/var/lib/cinder/volumes
        iscsi_protocol=iscsi
        volume_backend_name=lvm

    Examples:
        >>> type(conf)
        <class 'insights.parsers.cinder_conf.CinderConf'>
        >>> conf.sections()
        ['lvm']
        >>> 'lvm' in conf
        True
        >>> conf.has_option('DEFAULT', 'enabled_backends')
        True
        >>> conf.get("DEFAULT", "enabled_backends")
        'tripleo_ceph'
        >>> conf.get("DEFAULT", "glance_api_ssl_compression")
        'False'
        >>> conf.getboolean("DEFAULT", "glance_api_ssl_compression")
        False
        >>> conf.getint("DEFAULT", "osapi_volume_workers")
        32
    """
    pass



"""
Module responsible for scheduling Insights data collection in cron
"""
from __future__ import absolute_import
import os
import logging

from .constants import InsightsConstants as constants
from .utilities import run_command_get_output

APP_NAME = constants.app_name
logger = logging.getLogger(__name__)


def cron_source(config):
    return '/etc/%s/%s%s.cron' % (APP_NAME, APP_NAME, '')


class InsightsSchedulerCron(object):

    def __init__(self, config, source=None, target='/etc/cron.daily/' + APP_NAME):
        self.target = target
        self.source = source if source else cron_source(config)

    @property
    def active(self):
        if os.path.exists(self.source):
            return os.path.isfile(self.target)
        return False

    def schedule(self):
        logger.debug('Scheduling cron.daily')
        try:
            if not os.path.exists(self.target):
                os.symlink(self.source, self.target)
            return True
        except OSError:
            logger.exception('Could not link cron.daily')
            return False

    def remove_scheduling(self):
        logger.debug('Removing all cron tasks')
        try:
            # Remove symlinks in the case of rhel 6 running cron jobs
            if os.path.exists(self.target):
                os.remove(self.target)
            return True
        except OSError:
            logger.exception('Could not remove cron.daily')
            return False


class InsightsSchedulerSystemd(object):
    ALL_TIMERS = ("insights-client", "insights-client-checkin")

    def __init__(self):
        """
        Looks for loaded timers using `systemctl show`, stores their names in self.loaded_timers. No loaded timers
        produce (). If an error occurs, self.loaded_timers becomes None and all methods (schedule, remove_scheduling,
        active) then return None.
        """
        results = self._run_systemctl_commands(self.ALL_TIMERS, "show", "--property", "LoadState")
        if not results:
            self.loaded_timers = None  # Command failed.
        else:
            self.loaded_timers = tuple(
                timer
                for timer, result in results.items()
                if result["status"] == 0 and result["output"] == "LoadState=loaded\n"
            )
            if not self.loaded_timers:
                logger.warning("No loaded timers found")

    @staticmethod
    def _run_systemctl_command(*args):
        cmd_args = " ".join(args)
        command = "systemctl %s" % cmd_args
        logger.debug("Running command %s", command)
        try:
            result = run_command_get_output(command)
        except OSError:
            logger.exception("Could not run %s", command)
            return None
        else:
            logger.debug("Status: %s", result["status"])
            logger.debug("Output: %s", result["output"])
            return result

    @classmethod
    def _run_systemctl_commands(cls, timers, *args):
        if timers is None:
            return None  # Could not list loaded timers on init.

        results = {}

        for timer in timers:
            unit = "%s.timer" % timer
            command_args = args + (unit,)
            result = cls._run_systemctl_command(*command_args)
            if not result:
                return None  # Command failed.
            results[timer] = result

        return results

    @property
    def active(self):
        """
        Runs systemctl is-enabled for each loaded timers. Returns True if all loaded timers are enabled, None if any
        systemctl command fails - here or in init.
        """
        results = self._run_systemctl_commands(self.loaded_timers, "is-enabled")
        return results and all(result["status"] == 0 for result in results.values())

    def schedule(self):
        """
        Runs systemctl enable --now for each loaded timers. Returns True if all loaded timers are successfully enabled,
        False if any of them remains inactive. If no timer is loaded, returns {}, if any systemctl command fails - here
        or in init - returns None. Both falsey as nothing has been actually enabled.
        """
        logger.debug("Starting systemd timers")
        results = self._run_systemctl_commands(self.loaded_timers, "enable", "--now")
        return results and self.active

    def remove_scheduling(self):
        """
        Runs systemctl disable --now for each loaded timers. Returns True if all loaded timers are successfully
        disabled, False if any of them remains active. If no timer is loaded, returns {}, if any systemctl command
        fails - here or in init - returns None. Both falsey as nothing has been actually disabled.
        """
        logger.debug("Stopping all systemd timers")
        results = self._run_systemctl_commands(self.loaded_timers, "disable", "--now")

        if results:
            active = self.active
            return None if active is None else not active
        else:
            return results


def get_scheduler(config, source=None, target='/etc/cron.daily/' + APP_NAME):
    source = source if source else cron_source(config)
    if os.path.exists(source):
        return InsightsSchedulerCron(config, source, target)
    else:
        return InsightsSchedulerSystemd()



"""
PcsQuorumStatus - Commands ``pcs quorum status``
================================================
"""
from insights import parser, CommandParser
from insights.parsers import SkipException, ParseException, parse_fixed_table
from insights.specs import Specs


@parser(Specs.pcs_quorum_status)
class PcsQuorumStatus(CommandParser):
    """
    Class for parsing the output of `pcs quorum status` command.

    Typical output of the command is::

        Quorum information
        ------------------
        Date:             Wed Jun 29 13:17:02 2016
        Quorum provider:  corosync_votequorum
        Nodes:            2
        Node ID:          1
        Ring ID:          1/8272
        Quorate:          Yes

        Votequorum information
        ----------------------
        Expected votes:   3
        Highest expected: 3
        Total votes:      3
        Quorum:           2
        Flags:            Quorate Qdevice

        Membership information
        ----------------------
            Nodeid      Votes    Qdevice Name
                 1          1    A,V,NMW node1 (local)
                 2          1    A,V,NMW node2
                 0          1            Qdevice

    Attributes:
        quorum_info (dict): Dicts where keys are the feature name of quorum information and
            values are the corresponding feature value.
        votequorum_info (dict): Dicts where keys are the feature name of votequorum information and
            values are the corresponding feature value.
        membership_info (list): List of dicts where keys are the feature name of each node and
            values are the corresponding feature value.
    Raises:
        SkipException: When input is empty.
        ParseException: When input cannot be parsed.

    Examples:
        >>> type(pcs_quorum_status)
        <class 'insights.parsers.pcs_quorum_status.PcsQuorumStatus'>
        >>> pcs_quorum_status.quorum_info['Node ID']
        '1'
        >>> pcs_quorum_status.votequorum_info['Expected votes']
        '3'
        >>> pcs_quorum_status.membership_info[0]['Name']
        'node1 (local)'
    """

    def parse_content(self, content):
        if not content:
            raise SkipException("Empty content")
        if len(content) < 21 or not ('Quorum information' in content[0] and
                                     'Votequorum information' in content[9] and
                                     'Membership information' in content[17]):
            raise ParseException("Incorrect content: '{0}'".format(content))

        self.quorum_info = {}
        for line in content[2:7]:
            key, value = line.split(':', 1)
            self.quorum_info[key.strip()] = value.strip()

        self.votequorum_info = {}
        for line in content[11:15]:
            key, value = line.split(':', 1)
            self.votequorum_info[key.strip()] = value.strip()

        self.membership_info = parse_fixed_table(content[19:])



import errno
import hashlib
import os

from insights.util import subproc


def read_in_chunks(file_object, chunk_size=1024):
    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data


def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)


def hash_bytestr_iter(bytesiter, hasher):
    for block in bytesiter:
        hasher.update(block)
    return hasher.hexdigest()


def file_as_blockiter(afile, blocksize=65536):
    block = afile.read(blocksize)
    while len(block) > 0:
        yield block
        block = afile.read(blocksize)


def sha256(path):
    with open(path, "rb") as f:
        block_iter = file_as_blockiter(f)
        hasher = hashlib.sha256()
        return hash_bytestr_iter(block_iter, hasher)


def remove(path, chmod=False):
    """Remove a file or directory located on the filesystem at path.

    If chmod is True, chmod -R 755 is executed on the path
    before rm -rf path is called.

    Parameters
    ----------
    path : str
        file system path to an existing file or directory
    chmod : bool
        If True, chmod -R 755 is executed on path before
        it's removed.

    Raises
    ------
    CalledProcessError
        If any part of the removal process fails.
    """

    if not os.path.exists(path):
        return

    if chmod:
        cmd = "chmod -R 755 %s" % path
        subproc.call(cmd)

    cmd = 'rm -rf "{p}"'.format(p=path)
    subproc.call(cmd)


def ensure_path(path, mode=0o755):
    """Ensure that path exists in a multiprocessing safe way.

    If the path does not exist, recursively create it and its parent
    directories using the provided mode.  If the path already exists,
    do nothing.  The umask is cleared to enable the mode to be set,
    and then reset to the original value after the mode is set.

    Parameters
    ----------
    path : str
        file system path to a non-existent directory
        that should be created.
    mode : int
        octal representation of the mode to use when creating
        the directory.

    Raises
    ------
    OSError
        If os.makedirs raises an OSError for any reason
        other than if the directory already exists.
    """

    if path:
        umask = os.umask(0o022)
        try:
            os.makedirs(path, mode)
            os.umask(umask)
        except OSError as e:
            os.umask(umask)
            if e.errno != errno.EEXIST:
                raise


def size(path):
    """Return the size of the file.

    Parameters
    ----------
    path : str
        absolute path to a file.

    Returns
    -------
    int
        size of the file in bytes.
    """

    return os.stat(path).st_size



from insights.parsers.mariadb_log import MariaDBLog
from insights.tests import context_wrap

from datetime import datetime

MARIADB_LOG = """
161109  9:25:42 [Warning] SSL error: SSL_CTX_set_default_verify_paths failed
161109  9:25:42 [Note] WSREP: Service disconnected.
161109  9:25:43 [Note] WSREP: Some threads may fail to exit.
161109 14:28:24 InnoDB: Initializing buffer pool, size = 128.0M
161109 14:28:24 InnoDB: Completed initialization of buffer pool
"""


def test_mariadb_log():
    log = MariaDBLog(context_wrap(MARIADB_LOG))
    assert len(log.get("[Warning]")) == 1
    assert len(log.get("[Note]")) == 2
    assert 'SSL_CTX_set_default_verify_paths' in log
    assert len(list(log.get_after(datetime(2016, 11, 9, 14, 0, 0)))) == 2
